import keyword
import logging
import re

NON_WORD = re.compile(r'[\W]+')
LEADING_NUM = re.compile(r'^\d+')
LOGGING_LEVEL_MAP = {
    'critical': logging.CRITICAL,
    'fatal': logging.FATAL,
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'warn': logging.WARN,
    'info': logging.INFO,
    'debug': logging.DEBUG,
    'notset': logging.NOTSET,
}


def to_cmd_name(name: str) -> str:
    return NON_WORD.sub('-', name).strip('-').lower()


def to_variable_name(string: str) -> str:
    # Replace any non-word characters with underscore
    string = NON_WORD.sub('_', string).strip()
    # Remove leading digits
    string = LEADING_NUM.sub('', string)
    # Convert to lowercase
    string = string.lower()
    if keyword.iskeyword(string):
        return string + '_'
    return string


def to_boolean(string: str) -> bool:
    if string.lower() in ['true', '1', 'yes', 'y', 'active']:
        return True
    if string.lower() in ['false', '0', 'no', 'n', 'inactive']:
        return False
    raise Exception(f'Cannot infer boolean value from "{string}"')


def to_logging_level(logging_level_str: str) -> int:
    lower_logging_level_str = logging_level_str.lower()
    if lower_logging_level_str in LOGGING_LEVEL_MAP:
        return LOGGING_LEVEL_MAP[lower_logging_level_str]
    return logging.WARNING



import aiofiles


async def read_text_file_async(file_name: str) -> str:
    async with aiofiles.open(file_name, mode='r') as file:
        content = await file.read()
    return content


async def write_text_file_async(file_name: str, content: str):
    async with aiofiles.open(file_name, mode='w') as file:
        await file.write(content)


async def append_text_file_async(file_name: str, additional_content: str):
    content = await read_text_file_async(file_name)
    if content != '':
        additional_content = '\n' + additional_content
    new_content = content + additional_content
    await write_text_file_async(file_name, new_content)



from typing import Mapping
from ..._group import project_group
from ....task.cmd_task import CmdTask
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from ....config.config import version
from .._common.input import project_dir_input, project_name_input
from ..project_task.task_factory import create_add_project_automation_task

import os

# Common definitions

current_dir = os.path.dirname(__file__)


def copy_resource_replacement_mutator(
    task: ResourceMaker, replacements: Mapping[str, str]
) -> Mapping[str, str]:
    replacements['baseProjectDir'] = os.path.basename(
        replacements.get('projectDir', '')
    )
    if replacements.get('projectName', '') == '':
        replacements['projectName'] = replacements.get('baseProjectDir', '')
    return replacements


# Task definitions

copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        project_name_input
    ],
    replacements={
        'projectDir': '{{input.project_dir}}',
        'projectName': '{{input.project_name}}',
        'zrbVersion': version,
    },
    replacement_mutator=copy_resource_replacement_mutator,
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{input.project_dir}}',
    locks=['{{input.project_dir}}/zrb_init.py'],
    excludes=[
        '*/__pycache__',
    ]
)

add_project_task = create_add_project_automation_task(
    upstreams=[copy_resource]
)

create_project = CmdTask(
    name='create',
    group=project_group,
    upstreams=[add_project_task],
    inputs=[project_dir_input],
    cmd=[
        'set -e',
        'cd "{{input.project_dir}}"',
        'if [ ! -d .git ]',
        'then',
        '  echo Initialize project git repository',
        '  git init',
        'fi',
        'echo "Happy coding :)"',
    ]
)
runner.register(create_project)



from typing import AsyncIterator
from httpx import AsyncClient
from src.config import app_name
import pytest


@pytest.mark.asyncio
async def test_get_liveness(
    test_client_generator: AsyncIterator[AsyncClient]
):
    async for client in test_client_generator:
        response = await client.get("/liveness")
        assert response.status_code == 200
        assert response.json() == {'app': app_name, 'alive': True}


@pytest.mark.asyncio
async def test_head_liveness(
    test_client_generator: AsyncIterator[AsyncClient]
):
    async for client in test_client_generator:
        response = await client.head("/liveness")
        assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_readiness(
    test_client_generator: AsyncIterator[AsyncClient]
):
    async for client in test_client_generator:
        response = await client.get("/readiness")
        assert response.status_code == 200
        assert response.json() == {'app': app_name, 'ready': True}


@pytest.mark.asyncio
async def test_head_readiness(
    test_client_generator: AsyncIterator[AsyncClient]
):
    async for client in test_client_generator:
        response = await client.head("/readiness")
        assert response.status_code == 200



from typing import Mapping, List
from dotenv import dotenv_values
import copy
import jsons
import os
import re

NON_ALPHA_NUM = re.compile(r'[^a-zA-Z0-9]+')
CURRENT_DIR: str = os.path.dirname(__file__)
MODE: str = os.getenv('MODE', 'monolith')
NAMESPACE: str = os.getenv('NAMESPACE', 'default')

MODULE_JSON_STR: str = os.getenv('MODULES', '[]')
MODULES: List[str] = jsons.loads(MODULE_JSON_STR)

APP_DIR: str = os.path.abspath(os.path.join(CURRENT_DIR, '..', 'src'))
TEMPLATE_ENV_FILE_NAME: str = os.path.join(APP_DIR, 'template.env')
TEMPLATE_ENV_MAP: Mapping[str, str] = {
    key: os.getenv(key, default_value)
    for key, default_value in dotenv_values(TEMPLATE_ENV_FILE_NAME).items()
}

BROKER_TYPE: str = os.getenv('BROKER_TYPE', '')
if BROKER_TYPE == '':
    BROKER_TYPE = TEMPLATE_ENV_MAP.get('APP_BROKER_TYPE', 'rabbitmq')

RABBITMQ_AUTH_USERNAME: str = os.getenv('RABBITMQ_AUTH_USERNAME', 'root')
RABBITMQ_AUTH_PASSWORD: str = os.getenv('RABBITMQ_AUTH_PASSWORD', 'toor')

REDPANDA_AUTH_SASL_ENABLED: bool = os.getenv('REDPANDA_AUTH_SASL_ENABLED', 'true').lower() == 'true'  # noqa
REDPANDA_AUTH_MECHANISM: str = os.getenv('REDPANDA_AUTH_MECHANISM', 'SCRAM-SHA-512')  # noqa
REDPANDA_AUTH_USER_NAME: str = os.getenv('REDPANDA_AUTH_USER_NAME')
REDPANDA_AUTH_USER_PASSWORD: str = os.getenv('REDPANDA_AUTH_USER_PASSWORD')

POSTGRESQL_AUTH_POSTGRES_PASSWORD: str = os.getenv('POSTGRESQL_AUTH_POSTGRES_PASSWORD')  # noqa
POSTGRESQL_AUTH_USERNAME: str = os.getenv('POSTGRESQL_AUTH_USERNAME')
POSTGRESQL_AUTH_PASSWORD: str = os.getenv('POSTGRESQL_AUTH_PASSWORD')
POSTGRESQL_DB: str = os.getenv('POSTGRESQL_DB')


def get_app_monolith_env_map(
    template_env_map: Mapping[str, str], modules: List[str]
) -> Mapping[str, str]:
    env_map = copy.deepcopy(template_env_map)
    env_map['APP_RMQ_CONNECTION'] = f'amqp://{RABBITMQ_AUTH_USERNAME}:{RABBITMQ_AUTH_PASSWORD}@rabbitmq'  # noqa
    env_map['APP_KAFKA_BOOTSTRAP_SERVERS'] = 'redpanda:9092'
    env_map['APP_KAFKA_SASL_MECHANISM'] = REDPANDA_AUTH_MECHANISM
    env_map['APP_KAFKA_SASL_USER'] = REDPANDA_AUTH_USER_NAME
    env_map['APP_KAFKA_SASL_PASSWORD'] = REDPANDA_AUTH_USER_PASSWORD
    env_map['APP_DB_CONNECTION'] = f'postgresql+psycopg2://{POSTGRESQL_AUTH_USERNAME}:{POSTGRESQL_AUTH_PASSWORD}@postgresql:5432/{POSTGRESQL_DB}'  # noqa
    for module_name in modules:
        env_name = get_module_flag_env_name(module_name)
        env_map[env_name] = 'true'
    env_map['APP_ENABLE_MESSAGE_CONSUMER'] = 'true'
    env_map['APP_ENABLE_RPC_SERVER'] = 'true'
    env_map['APP_ENABLE_FRONTEND'] = 'true'
    env_map['APP_ENABLE_API'] = 'true'
    return env_map


def get_app_gateway_env_map(
    template_env_map: Mapping[str, str], modules: List[str]
) -> Mapping[str, str]:
    env_map = get_app_monolith_env_map(template_env_map, modules)
    for module_name in modules:
        env_name = get_module_flag_env_name(module_name)
        env_map[env_name] = 'true'
    env_map['APP_ENABLE_MESSAGE_CONSUMER'] = 'false'
    env_map['APP_ENABLE_RPC_SERVER'] = 'false'
    env_map['APP_ENABLE_FRONTEND'] = 'true'
    env_map['APP_ENABLE_API'] = 'true'
    env_map['APP_DB_AUTO_MIGRATE'] = 'false'
    return env_map


def get_app_service_env_map(
    template_env_map: Mapping[str, str],
    modules: List[str],
    current_module: str
) -> Mapping[str, str]:
    env_map = get_app_monolith_env_map(template_env_map, modules)
    for module_name in modules:
        env_name = get_module_flag_env_name(module_name)
        if module_name == current_module:
            env_map[env_name] = 'true'
            continue
        env_map[env_name] = 'false'
    env_map['APP_ENABLE_MESSAGE_CONSUMER'] = 'true'
    env_map['APP_ENABLE_RPC_SERVER'] = 'true'
    env_map['APP_ENABLE_FRONTEND'] = 'false'
    env_map['APP_ENABLE_API'] = 'false'
    return env_map


def get_module_flag_env_name(module_name: str) -> str:
    upper_snake_module_name = to_snake_case(module_name).upper()
    return f'APP_ENABLE_{upper_snake_module_name}_MODULE'


def to_kebab_case(text: str) -> str:
    text = to_alphanum(text)
    return '-'.join([
        x.lower() for x in text.split(' ')
    ])


def to_snake_case(text: str) -> str:
    text = to_alphanum(text)
    return '_'.join([
        x.lower() for x in text.split(' ')
    ])


def to_alphanum(text: str) -> str:
    return NON_ALPHA_NUM.sub(' ', text)



from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller, Server
from module.auth.entity.permission.rpc import (
    register_rpc as register_permission_rpc
)
from module.auth.entity.group.rpc import (
    register_rpc as register_group_rpc
)
from module.auth.entity.user.rpc import (
    register_rpc as register_user_rpc
)


def register_rpc(
    logger: Logger,
    rpc_server: Server,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "auth"')
    register_permission_rpc(logger, rpc_server, rpc_caller, publisher)
    register_group_rpc(logger, rpc_server, rpc_caller, publisher)
    register_user_rpc(logger, rpc_server, rpc_caller, publisher)



from zrb_init import concat, register_trainer, fibo

print('Concat result', concat.create_main_loop()(
    'angry', 'bulbasaur', separator='-'
))

print('Register trainer result', register_trainer.create_main_loop()(
    name='John',
    password='secret',
    age=10,
    starter_pokemon='bulbasaur'
))

print('Fibo result', fibo.create_main_loop()(
    n=7
))

print('Fibo result (without default value)', fibo.create_main_loop()())



from typing import List, Optional
from typeguard import typechecked
from dotenv import dotenv_values
from .env import Env


@typechecked
class EnvFile():

    def __init__(
        self,
        env_file: str,
        prefix: Optional[str] = None,
        renderable: bool = False
    ):
        self.env_file = env_file
        self.prefix = prefix.upper() if prefix is not None else None
        self.renderable = renderable

    def get_envs(self) -> List[Env]:
        envs: List[Env] = []
        env_map = dotenv_values(self.env_file)
        for key, value in env_map.items():
            os_name: Optional[str] = None
            if self.prefix is not None and self.prefix != '':
                os_name = f'{self.prefix}_{key}'
            envs.append(Env(
                name=key,
                os_name=os_name,
                default=value,
                renderable=self.renderable
            ))
        return envs



from component.messagebus import publisher
from module.auth.component.repo.permission_repo import permission_repo
from module.auth.entity.permission.model import PermissionModel

permission_model: PermissionModel = PermissionModel(permission_repo, publisher)



from typing import Any, Mapping, List, Optional
from core.messagebus.messagebus import Admin
import logging
import aiormq
import asyncio


class RMQEventConfig():
    def __init__(
        self,
        queue_name: str,
        exchange_name: str = '',
    ):
        self.queue_name = queue_name
        self.exchange_name = exchange_name


class RMQAdmin(Admin):
    def __init__(
        self,
        logger: logging.Logger,
        configs: Mapping[str, RMQEventConfig],
        connection_string: str
    ):
        self.logger = logger
        self.connection_string = connection_string
        self.configs = configs
        self._existing_events: Mapping[str, bool] = {}

    async def create_events(self, event_names: List[str]):
        # Only handle non-existing events
        event_names = [
            event_name for event_name in event_names
            if event_name not in self._existing_events
        ]
        if len(event_names) == 0:
            return
        try:
            connection = await aiormq.connect(self.connection_string)
            channel = await connection.channel()
            for event_name in event_names:
                config = self.get_config(event_name)
                if config.exchange_name != '':
                    await self._declare_fanned_out_exchange(channel, config)
                    self._existing_events[event_name] = True
                    continue
                await self._declare_queue(channel, config)
                self._existing_events[event_name] = True
            await self._clean_up(connection, channel)
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error(
                ' '.join([
                    '🐰 [rabbitmq-admin] Something wrong when ',
                    f'creating events: {event_names}'
                ]),
                exc_info=True
            )

    async def delete_events(self, event_names: List[str]):
        # Only handle existing events
        event_names = [
            event_name for event_name in event_names
            if event_name in self._existing_events
        ]
        if len(event_names) == 0:
            return
        try:
            connection = await aiormq.connect(self.connection_string)
            channel = await connection.channel()
            for event_name in event_names:
                config = self.get_config(event_name)
                # delete the queue
                await channel.queue_delete(queue=config.queue_name)
                if config.exchange_name != '':
                    # delete the exchange
                    await channel.exchange_delete(
                        exchange_name=config.exchange_name
                    )
                del self._existing_events[event_name]
            await self._clean_up(connection, channel)
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error(
                ' '.join([
                    '🐰 [rabbitmq-admin] Something wrong when ',
                    f'deleting events: {event_names}'
                ]),
                exc_info=True
            )

    async def _clean_up(
        self, connection: aiormq.Connection, channel: aiormq.Channel
    ):
        try:
            await channel.close()
            await connection.close()
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error('🐰 [rabbitmq-admin]', exc_info=True)

    async def _declare_fanned_out_exchange(
        self, channel: aiormq.Channel, config: RMQEventConfig
    ):
        await channel.exchange_declare(
            exchange=config.exchange_name,
            exchange_type='fanout',
            durable=False,
            auto_delete=False
        )
        # declare a queue
        result = await self._declare_queue(channel, config)
        queue_name = result['queue']
        # bind the queue to the exchange
        await channel.queue_bind(
            queue=queue_name,
            exchange=config.exchange_name,
            routing_key=''
        )

    async def _declare_queue(
        self, channel: aiormq.Channel, config: RMQEventConfig
    ) -> Any:
        return await channel.queue_declare(
            queue=config.queue_name,
            durable=True,
            auto_delete=False
        )

    def get_config(self, event_name: str) -> RMQEventConfig:
        if event_name in self.configs:
            return self.configs[event_name]
        return RMQEventConfig(queue_name=event_name)

    def get_exchange_name(self, event_name: str) -> str:
        event_config = self.get_config(event_name)
        return event_config.exchange_name

    def get_queue_name(self, event_name: str) -> str:
        event_config = self.get_config(event_name)
        return event_config.queue_name


def must_get_rmq_admin(
    logger: logging.Logger,
    rmq_admin: Optional[RMQAdmin],
    connection_string: str
) -> RMQAdmin:
    if rmq_admin is None:
        return RMQAdmin(
            logger=logger,
            configs={},
            connection_string=connection_string
        )
    return rmq_admin



from typing import Any
from ._group import principle_show_group
from ..task.decorator import python_task
from ..task.task import Task
from ..runner import runner


@python_task(
    name='solid',
    group=principle_show_group,
    description='Show SOLID principle',
    runner=runner
)
async def show_solid_principle(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('S - Single Responsibility Principle')
    task.print_out('O - Open/Closed Principle')
    task.print_out('L - Liskov’s Substitution Principle')
    task.print_out('I - Interface Segregation Principle')
    task.print_out('D - Dependency Inversion Principle')


@python_task(
    name='yagni',
    group=principle_show_group,
    description='Show YAGNI principle',
    runner=runner
)
async def show_yagni_principle(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('Y - You')
    task.print_out('A - Aren\'t')
    task.print_out('G - Gonna')
    task.print_out('N - Need')
    task.print_out('I - It')


@python_task(
    name='dry',
    group=principle_show_group,
    description='Show dry principle',
    runner=runner
)
async def show_dry_principle(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('D - Don\'t')
    task.print_out('R - Repeat')
    task.print_out('Y - Yourself')


@python_task(
    name='kiss',
    group=principle_show_group,
    description='Show kiss principle',
    runner=runner
)
async def show_kiss_principle(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('K - Keep')
    task.print_out('I - It')
    task.print_out('S - Simple')
    task.print_out('S - Stupid')



from typing import Any, List, Mapping
from abc import ABC, abstractmethod
from sqlalchemy import Column, String, or_
from sqlalchemy.orm import Session, relationship
from sqlalchemy.engine import Engine
from core.repo import Repo, DBEntityMixin, DBRepo
from module.auth.schema.user import User, UserData, UserLogin
from module.auth.core import PasswordHasher
from module.auth.component import Base
from module.auth.entity.table import user_group, user_permission
from module.auth.entity.permission.repo import DBEntityPermission
from module.auth.entity.group.repo import DBEntityGroup

import logging


class DBEntityUser(Base, DBEntityMixin):
    __tablename__ = "users"
    username = Column(String)
    phone = Column(String)
    email = Column(String)
    description = Column(String)
    hashed_password = Column(String)
    permissions = relationship(
        "DBEntityPermission",
        secondary=user_permission
    )
    groups = relationship(
        "DBEntityGroup",
        secondary=user_group
    )


class UserRepo(Repo[User, UserData], ABC):
    @abstractmethod
    async def get_by_user_login(self, user_login: UserLogin) -> User:
        pass


class UserDBRepo(
    DBRepo[User, UserData], UserRepo
):
    schema_cls = User
    db_entity_cls = DBEntityUser

    def __init__(
        self,
        logger: logging.Logger,
        engine: Engine,
        password_hasher: PasswordHasher
    ):
        super().__init__(logger, engine)
        self.password_hasher = password_hasher

    async def get_by_user_login(self, user_login: UserLogin) -> User:
        db = self._create_db_session()
        db_users: List[DBEntityUser] = []
        try:
            search_filter = or_(
                DBEntityUser.username == user_login.identity,
                DBEntityUser.phone == user_login.identity,
                DBEntityUser.email == user_login.identity
            )
            db_users = [
                db_user
                for db_user in self._get_by_criterion(db, search_filter)
                if self.password_hasher.check_password(
                    user_login.password, db_user.hashed_password
                )
            ]
            if len(db_users) == 0:
                raise ValueError('Not found: Cannot find any user')
            return self._db_entity_to_schema(db, db_users[0])
        finally:
            db.close()

    def _schema_data_to_db_entity_map(
        self, db: Session, user_data: UserData
    ) -> Mapping[str, Any]:
        db_entity_map = super()._schema_data_to_db_entity_map(db, user_data)
        # Transform permissions
        db_entity_map['permissions'] = db.query(DBEntityPermission).filter(
            DBEntityPermission.id.in_(user_data.permissions)
        ).all()
        # Transform groups
        db_entity_map['groups'] = db.query(DBEntityGroup).filter(
            DBEntityGroup.id.in_(user_data.groups)
        ).all()
        # add hashed password if necessary
        if user_data.password != '':
            db_entity_map['hashed_password'] = self.password_hasher.hash_password( # noqa
                user_data.password
            )
        return db_entity_map



from ..helper.string.conversion import to_boolean, to_logging_level
import os
import pkg_resources


def get_version():
    try:
        dist = pkg_resources.get_distribution("zrb")
        return dist.version
    except pkg_resources.DistributionNotFound:
        import flit
        meta = flit.read_module_metadata("zrb")
        return meta["module"]["version"]


default_shell = os.getenv('ZRB_SHELL', 'bash')
init_scripts = os.getenv('ZRB_INIT_SCRIPTS', '').split(':')
logging_level = to_logging_level(os.getenv('ZRB_LOGGING_LEVEL', 'WARNING'))
should_load_builtin = to_boolean(os.getenv('ZRB_SHOULD_LOAD_BUILTIN', '1'))
env_prefix = os.getenv('ZRB_ENV', '')
show_advertisement = to_boolean(os.getenv('ZRB_SHOW_ADVERTISEMENT', '1'))
show_prompt = to_boolean(os.getenv('ZRB_SHOW_PROMPT', '1'))
version = get_version()



from component.log import logger
from component.db_connection import engine
from module.auth.entity.permission.repo import (
    PermissionRepo, PermissionDBRepo
)

permission_repo: PermissionRepo = PermissionDBRepo(
    logger=logger, engine=engine
)



from zrb import BoolInput, IntInput, StrInput, Env
import os

###############################################################################
# Constants
###############################################################################

CURRENT_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, '..', '..'))
RESOURCE_DIR = os.path.join(PROJECT_DIR, 'src', 'kebab-app-name')
DEPLOYMENT_DIR = os.path.join(RESOURCE_DIR, 'deployment')
APP_DIR = os.path.join(RESOURCE_DIR, 'src')
TEMPLATE_ENV_FILE_NAME = os.path.join(APP_DIR, 'template.env')

###############################################################################
# Input Definitions
###############################################################################

local_input = BoolInput(
    name='local-kebab-app-name',
    description='Use "kebab-app-name" on local machine',
    prompt='Use "kebab-app-name" on local machine?',
    default=True
)

https_input = BoolInput(
    name='kebab-app-name-https',
    description='Whether "kebab-app-name" run on HTTPS',
    prompt='Is "kebab-app-name" run on HTTPS?',
    default=False
)

host_input = StrInput(
    name='kebab-app-name-host',
    description='Hostname of "kebab-app-name"',
    prompt='Hostname of "kebab-app-name"',
    default='localhost'
)

image_input = StrInput(
    name='kebab-app-name-image',
    description='Image name of "kebab-app-name"',
    prompt='Image name of "kebab-app-name"',
    default='app-image-name:latest'
)

replica_input = IntInput(
    name='kebab-app-name-replica',
    description='Replica of "kebab-app-name"',
    prompt='Replica of "kebab-app-name"',
    default=1,
)

pulumi_stack_input = StrInput(
    name='kebab-app-name-pulumi-stack',
    description='Pulumi stack name for "kebab-app-name"',
    prompt='Pulumi stack name for "kebab-app-name"',
    default=os.getenv('ZRB_ENV', 'dev')
)

###############################################################################
# Env Definitions
###############################################################################

image_env = Env(
    name='IMAGE',
    os_name='CONTAINER_ENV_PREFIX_IMAGE',
    default='{{input.snake_app_name_image}}'
)

pulumi_backend_url_env = Env(
    name='PULUMI_BACKEND_URL',
    os_name='PULUMI_ENV_PREFIX_BACKEND_URL',
    default=f'file://{DEPLOYMENT_DIR}/state'
)

pulumi_config_passphrase_env = Env(
    name='PULUMI_CONFIG_PASSPHRASE',
    os_name='PULUMI_ENV_PREFIX_CONFIG_PASSPHRASE',
    default='secret'
)

deployment_replica_env = Env(
    name='REPLICA',
    os_name='DEPLOYMENT_ENV_PREFIX',
    default='{{input.snake_app_name_replica}}'
)



from zrb.builtin.base64 import encode, decode


def test_encode():
    main_loop = encode.create_main_loop()
    result = main_loop(text='Philosopher Stone')
    assert result == 'UGhpbG9zb3BoZXIgU3RvbmU='


def test_decode():
    main_loop = decode.create_main_loop()
    result = main_loop(text='UGhpbG9zb3BoZXIgU3RvbmU=')
    assert result == 'Philosopher Stone'



from zrb import (
    BoolInput, ChoiceInput, StrInput, Env, EnvFile,
    HTTPChecker, PortChecker
)
import jsons
import os

###############################################################################
# Constants
###############################################################################

CURRENT_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, '..', '..'))
RESOURCE_DIR = os.path.join(PROJECT_DIR, 'src', 'kebab-app-name')
DEPLOYMENT_DIR = os.path.join(RESOURCE_DIR, 'deployment')
DEPLOYMENT_TEMPLATE_ENV_FILE_NAME = os.path.join(
    DEPLOYMENT_DIR, 'template.env'
)
APP_DIR = os.path.join(RESOURCE_DIR, 'src')
APP_FRONTEND_DIR = os.path.join(APP_DIR, 'frontend')
APP_FRONTEND_BUILD_DIR = os.path.join(APP_FRONTEND_DIR, 'build')
APP_TEMPLATE_ENV_FILE_NAME = os.path.join(APP_DIR, 'template.env')
SKIP_CONTAINER_EXECUTION = '{{not input.local_snake_app_name}}'
SKIP_SUPPORT_CONTAINER_EXECUTION = ' '.join([
    '{{',
    'not input.local_snake_app_name',
    'or env.get("APP_BROKER_TYPE") not in ["rabbitmq", "kafka"]',
    '}}',
])
SKIP_LOCAL_MONOLITH_EXECUTION = ' '.join([
    '{{',
    'not input.local_snake_app_name',
    'or input.snake_app_name_mode == "microservices"',
    '}}',
])
SKIP_LOCAL_MICROSERVICES_EXECUTION = ' '.join([
    '{{',
    'not input.local_snake_app_name',
    'or input.snake_app_name_mode == "monolith"',
    '}}',
])

MODULE_CONFIG_PATH = os.path.join(CURRENT_DIR, 'config', 'modules.json')
with open(MODULE_CONFIG_PATH) as file:
    MODULE_JSON_STR = file.read()
MODULES = jsons.loads(MODULE_JSON_STR)


###############################################################################
# Checker Task Definitions
###############################################################################

rabbitmq_management_checker = HTTPChecker(
    name='check-rabbitmq-management',
    port='{{env.get("RABBITMQ_MANAGEMENT_HOST_PORT", "15672")}}',
    is_https='{{input.snake_app_name_https}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "rabbitmq"}}'
)

rabbitmq_checker = PortChecker(
    name='check-rabbitmq',
    port='{{env.get("RABBITMQ_HOST_PORT", "5672")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "rabbitmq"}}'
)

redpanda_console_checker = HTTPChecker(
    name='check-redpanda-console',
    method='GET',
    port='{{env.get("REDPANDA_CONSOLE_HOST_PORT", "9000")}}',
    is_https='{{input.snake_app_name_https}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

kafka_plaintext_checker = PortChecker(
    name='check-kafka-plaintext',
    port='{{env.get("KAFKA_PLAINTEXT_HOST_PORT", "29092")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

kafka_outside_checker = PortChecker(
    name='check-kafka-outside',
    port='{{env.get("KAFKA_OUTSIDE_HOST_PORT", "9092")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

pandaproxy_plaintext_checker = PortChecker(
    name='check-pandaproxy-plaintext',
    port='{{env.get("PANDAPROXY_PLAINTEXT_HOST_PORT", "29092")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

pandaproxy_outside_checker = PortChecker(
    name='check-pandaproxy-outside',
    port='{{env.get("PANDAPROXY_OUTSIDE_HOST_PORT", "9092")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

app_container_checker = HTTPChecker(
    name='check-kebab-app-name-container',
    host='{{input.snake_app_name_host}}',
    url='/readiness',
    port='{{env.get("HOST_PORT", "httpPort")}}',
    is_https='{{input.snake_app_name_https}}'
)

app_local_checker = HTTPChecker(
    name='check-kebab-app-name',
    host='{{input.snake_app_name_host}}',
    url='/readiness',
    port='{{env.APP_PORT}}',
    is_https='{{input.snake_app_name_https}}',
    skip_execution=SKIP_LOCAL_MICROSERVICES_EXECUTION
)

###############################################################################
# Input Definitions
###############################################################################

local_input = BoolInput(
    name='local-kebab-app-name',
    description='Use "kebab-app-name" on local machine',
    prompt='Use "kebab-app-name" on local machine?',
    default=True
)

mode_input = ChoiceInput(
    name='kebab-app-name-mode',
    description='"kebab-app-name" mode',
    prompt='Do you want to run "kebab-app-name" as monolith or microservices?',
    choices=['monolith', 'microservices'],
    default='monolith'
)

https_input = BoolInput(
    name='kebab-app-name-https',
    description='Whether "kebab-app-name" run on HTTPS',
    prompt='Is "kebab-app-name" run on HTTPS?',
    default=False
)

host_input = StrInput(
    name='kebab-app-name-host',
    description='Hostname of "kebab-app-name"',
    prompt='Hostname of "kebab-app-name"',
    default='localhost'
)

image_input = StrInput(
    name='kebab-app-name-image',
    description='Image name of "kebab-app-name"',
    prompt='Image name of "kebab-app-name"',
    default='app-image-name:latest'
)

pulumi_stack_input = StrInput(
    name='kebab-app-name-pulumi-stack',
    description='Pulumi stack name for "kebab-app-name"',
    prompt='Pulumi stack name for "kebab-app-name"',
    default=os.getenv('ZRB_ENV', 'dev')
)

###############################################################################
# Env file Definitions
###############################################################################

compose_env_file = EnvFile(
    env_file=os.path.join(CURRENT_DIR, 'config', 'docker-compose.env'),
    prefix='CONTAINER_ENV_PREFIX'
)
app_env_file = EnvFile(
    env_file=APP_TEMPLATE_ENV_FILE_NAME, prefix='ENV_PREFIX'
)

###############################################################################
# Env Definitions
###############################################################################

image_env = Env(
    name='IMAGE',
    os_name='CONTAINER_ENV_PREFIX_IMAGE',
    default='{{input.snake_app_name_image}}'
)

pulumi_backend_url_env = Env(
    name='PULUMI_BACKEND_URL',
    os_name='PULUMI_ENV_PREFIX_BACKEND_URL',
    default=f'file://{DEPLOYMENT_DIR}/state'
)

pulumi_config_passphrase_env = Env(
    name='PULUMI_CONFIG_PASSPHRASE',
    os_name='PULUMI_ENV_PREFIX_CONFIG_PASSPHRASE',
    default='secret'
)

local_app_port_env = Env(
    name='APP_PORT',
    os_name='ENV_PREFIX_APP_PORT',
    default='httpPort'
)

local_app_broker_type_env = Env(
    name='APP_BROKER_TYPE',
    os_name='ENV_PREFIX_APP_BROKER_TYPE',
    default='rabbitmq'
)



from zrb.helper.codemod.add_property_to_class import (
    add_property_to_class
)


def test_add_property_to_class():
    code = add_property_to_class(
        code='\n'.join([
            'class Cls():',
            '    existing_property: int',
            'class Something():',
            '    existing_property: int',
        ]),
        class_name='Something',
        property_name='some_property',
        property_type='int'
    )
    assert code == '\n'.join([
        'class Cls():',
        '    existing_property: int',
        'class Something():',
        '    existing_property: int',
        '    some_property: int',
    ])



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common.input import project_dir_input, task_name_input
from .._common.helper import validate_project_dir, create_register_task_module
from .._common.lock import new_task_lock
from ....helper import util

import os

# Common definitions

current_dir = os.path.dirname(__file__)


@python_task(
    name='validate',
    inputs=[
        project_dir_input,
        task_name_input,
    ]
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    task_name = kwargs.get('task_name')
    automation_file = os.path.join(
        project_dir, '_automate', f'{util.to_snake_case(task_name)}.py'
    )
    if os.path.exists(automation_file):
        raise Exception(f'Automation file already exists: {automation_file}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        task_name_input,
    ],
    upstreams=[validate],
    replacements={
        'taskName': '{{input.task_name}}',
    },
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    locks=[new_task_lock],
    excludes=[
        '*/__pycache__',
    ]
)

register_task_module = create_register_task_module(
    upstreams=[copy_resource]
)


@python_task(
    name='cmd-task',
    group=project_add_group,
    upstreams=[register_task_module],
    runner=runner
)
async def add_cmd_task(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from typing import Any, Mapping
from ._group import env_group
from ..helper.accessories.color import colored
from ..task.decorator import python_task
from ..task.task import Task
from ..runner import runner


@python_task(
    name='show',
    group=env_group,
    description='Show environment values',
    runner=runner
)
async def show(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    env_map = task.get_env_map()
    names = list(env_map.keys())
    names.sort()
    for name in names:
        value = env_map[name]
        colored_name = colored(name, color='green', attrs=['bold'])
        colored_equal = colored('=', color='grey', attrs=['dark'])
        colored_value = colored(value, attrs=['bold'])
        task.print_out(f'{colored_name}{colored_equal}{colored_value}')


@python_task(
    name='get',
    group=env_group,
    description='Get environment values',
    runner=runner
)
async def get(*args: Any, **kwargs: Any) -> Mapping[str, str]:
    task: Task = kwargs['_task']
    return task.get_env_map()



from typing import Annotated, List, Mapping
from fastapi import FastAPI, Depends
from fastapi.security import OAuth2PasswordRequestForm
from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller
from core.error import HTTPAPIException
from module.auth.core import Authorizer
from module.auth.schema.user import (
    User, UserData, UserResult, UserLogin
)
from module.auth.schema.token import AccessTokenData, TokenResponse
from module.auth.schema.request import RefreshTokenRequest, IsAuthorizedRequest
from module.auth.component import access_token_scheme, bearer_token_scheme


def register_auth_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register Login API for "auth.user"')

    @app.post('/api/v1/auth/login-oauth', response_model=TokenResponse)
    async def login_oauth(
        form_data: Annotated[OAuth2PasswordRequestForm, Depends()]
    ) -> TokenResponse:
        data = UserLogin(
            identity=form_data.username,
            password=form_data.password
        )
        return await _create_token(data=data)

    @app.post('/api/v1/auth/login', response_model=TokenResponse)
    async def login(data: UserLogin) -> TokenResponse:
        return await _create_token(data=data)

    async def _create_token(data: UserLogin) -> TokenResponse:
        try:
            token_response_dict = await rpc_caller.call(
                'auth_create_token', login_data=data.dict()
            )
            return TokenResponse(**token_response_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post('/api/v1/auth/refresh-token', response_model=TokenResponse)
    async def refresh_token(
        data: RefreshTokenRequest,
        refresh_token: str = Depends(bearer_token_scheme)
    ) -> TokenResponse:
        try:
            token_response_dict = await rpc_caller.call(
                'auth_refresh_token',
                refresh_token=refresh_token,
                access_token=data.access_token
            )
            return TokenResponse(**token_response_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post('/api/v1/auth/is-authorized', response_model=Mapping[str, bool])
    async def is_authorized(
        data: IsAuthorizedRequest,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ) -> Mapping[str, str]:
        try:
            user_id = user_token_data.user_id
            return await rpc_caller.call(
                'auth_is_user_authorized',
                id=user_id,
                permission_name=data.permission_names
            )
        except Exception as e:
            raise HTTPAPIException(error=e)


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register API for "auth.user"')

    @app.get(
        '/api/v1/auth/users', response_model=UserResult
    )
    async def get_users(
        keyword: str = '', limit: int = 100, offset: int = 0,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:get'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_get_user',
                keyword=keyword,
                criterion={},
                limit=limit,
                offset=offset,
                user_token_data=user_token_data.dict()
            )
            return UserResult(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.get(
        '/api/v1/auth/users/{id}', response_model=User
    )
    async def get_user_by_id(
        id: str,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:get_by_id'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_get_user_by_id',
                id=id, user_token_data=user_token_data.dict()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post(
        '/api/v1/auth/users', response_model=User
    )
    async def insert_user(
        data: UserData,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:insert'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_insert_user',
                data=data.dict(), user_token_data=user_token_data.dict()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.put(
        '/api/v1/auth/users/{id}', response_model=User
    )
    async def update_user(
        id: str, data: UserData,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:update'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_update_user',
                id=id, data=data.dict(), user_token_data=user_token_data.dict()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.delete(
        '/api/v1/auth/users/{id}', response_model=User
    )
    async def delete_user(
        id: str,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:delete'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_delete_user',
                id=id, user_token_data=user_token_data.dict()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)



from typing import Mapping, List
from _common import (
    NAMESPACE, MODULES, TEMPLATE_ENV_MAP, get_app_monolith_env_map,
    get_app_gateway_env_map, get_app_service_env_map, to_kebab_case,
    to_snake_case
)

import pulumi_kubernetes as k8s
import os

image = os.getenv('IMAGE', 'kebab-app-name:latest')
app_monolith_replica = int(os.getenv('REPLICA', '1'))
app_monolith_labels = {'app': 'kebab-app-name'}
app_gateway_replica = int(os.getenv('REPLICA_GATEWAY'))
app_gateway_labels = {'app': 'kebab-app-name-gateway'}

app_monolith_env_map = get_app_monolith_env_map(TEMPLATE_ENV_MAP, MODULES)
app_port = int(os.getenv(
    'APP_PORT', app_monolith_env_map.get('APP_PORT', '8080')
))
app_gateway_env_map = get_app_gateway_env_map(TEMPLATE_ENV_MAP, MODULES)


def create_app_microservices_deployments() -> List[k8s.apps.v1.Deployment]:
    deployments: List[k8s.apps.v1.Deployment] = []
    deployments.append(_create_app_deployment(
        resource_name='kebab-app-name-gateway',
        image=image,
        replica=app_gateway_replica,
        app_labels=app_gateway_labels,
        env_map=app_gateway_env_map,
        app_port=app_port
    ))
    for module in MODULES:
        kebab_module_name = to_kebab_case(module)
        snake_module_name = to_snake_case(module)
        upper_snake_module_name = snake_module_name.upper()
        app_service_replica = int(os.getenv(
            f'REPLICA_{upper_snake_module_name}_SERVICE', '1'
        ))
        app_service_env_map = get_app_service_env_map(
            TEMPLATE_ENV_MAP, MODULES, module
        )
        app_service_labels = {
            'app': f'kebab-app-name-{kebab_module_name}-service'
        }
        deployments.append(_create_app_deployment(
            resource_name=f'kebab-app-name-{kebab_module_name}-service',
            app_labels=app_service_labels,
            image=image,
            replica=app_service_replica,
            env_map=app_service_env_map,
            app_port=app_port
        ))
    return deployments


def create_app_microservices_services() -> List[k8s.core.v1.Service]:
    services: List[k8s.core.v1.Service] = []
    services.append(_create_app_service(
        resource_name='kebab-app-name-gateway',
        app_labels=app_gateway_labels,
        app_port=app_port,
        service_type='LoadBalancer',
        service_port=app_port,
        service_port_protocol='TCP'
    ))
    return services


def create_app_monolith_deployment() -> k8s.apps.v1.Deployment:
    return _create_app_deployment(
        resource_name='kebab-app-name',
        image=image,
        replica=app_monolith_replica,
        app_labels=app_monolith_labels,
        env_map=app_monolith_env_map,
        app_port=app_port
    )


def create_app_monolith_service() -> k8s.core.v1.Service:
    return _create_app_service(
        resource_name='kebab-app-name',
        app_labels=app_monolith_labels,
        app_port=app_port,
        service_type='LoadBalancer',
        service_port=app_port,
        service_port_protocol='TCP'
    )


def _create_app_deployment(
    resource_name: str,
    image: str,
    replica: int,
    app_labels: Mapping[str, str],
    env_map: Mapping[str, str],
    app_port: int
) -> k8s.apps.v1.Deployment:
    # Pulumi deployment docs:
    # https://www.pulumi.com/registry/packages/kubernetes/api-docs/apps/v1/deployment/
    deployment = k8s.apps.v1.Deployment(
        resource_name=resource_name,
        spec=k8s.apps.v1.DeploymentSpecArgs(
            selector=k8s.meta.v1.LabelSelectorArgs(match_labels=app_labels),
            replicas=replica,
            template=k8s.core.v1.PodTemplateSpecArgs(
                metadata=k8s.meta.v1.ObjectMetaArgs(
                    labels=app_labels,
                    namespace=NAMESPACE
                ),
                spec=k8s.core.v1.PodSpecArgs(
                    containers=[
                        k8s.core.v1.ContainerArgs(
                            name=resource_name,
                            image=image,
                            env=[
                                k8s.core.v1.EnvVarArgs(
                                    name=env_name,
                                    value=env_value
                                )
                                for env_name, env_value in env_map.items()
                            ],
                            ports=[
                                k8s.core.v1.ContainerPortArgs(
                                    container_port=app_port
                                )
                            ],
                            liveness_probe=k8s.core.v1.ProbeArgs(
                                http_get=k8s.core.v1.HTTPGetActionArgs(
                                    port=app_port,
                                    path='/liveness'
                                )
                            ),
                            readiness_probe=k8s.core.v1.ProbeArgs(
                                http_get=k8s.core.v1.HTTPGetActionArgs(
                                    port=app_port,
                                    path='/readiness'
                                )
                            ),
                        )
                    ]
                )
            )
        )
    )
    return deployment


def _create_app_service(
    resource_name: str,
    app_labels: Mapping[str, str],
    app_port: int,
    service_type: str = 'LoadBalancer',
    service_port: int = 8080,
    service_port_protocol: str = 'TCP',
) -> k8s.core.v1.Service:
    # Pulumi services docs:
    # https://www.pulumi.com/registry/packages/kubernetes/api-docs/core/v1/service/
    service = k8s.core.v1.Service(
        resource_name=resource_name,
        metadata=k8s.meta.v1.ObjectMetaArgs(
            namespace=NAMESPACE
        ),
        spec=k8s.core.v1.ServiceSpecArgs(
            selector=app_labels,
            ports=[
                k8s.core.v1.ServicePortArgs(
                    port=service_port,
                    protocol=service_port_protocol,
                    target_port=app_port,
                )
            ],
            type=service_type,
        )
    )
    return service



import random
import string


def get_random_name(
    separator: str = '-',
    add_random_digit: bool = True,
    digit_count: int = 3
) -> str:
    prefixes = [
        "aurum", "argentum", "platinum", "mercurius", "sulfur", "sal",
        "luna", "sol", "ferrum", "cuprum", "argent", "aurora", "citrin",
        "coral", "diamond", "dragon", "emerald", "garnet", "jade", "onyx",
        "opal", "pearl", "ruby", "sapphire", "topaz", "turquoise", "verde",
        "zircon"
    ]
    suffixes = [
        "philosophorum", "spiritus", "tinctura", "essentia", "elixir",
        "praeparatum", "aether", "vitae", "lapis", "metallum", "aureum",
        "caelestis", "chrysopoeia", "cosmicum", "deum", "draconis",
        "elementorum", "hermetica", "illuminationis", "magnum", "mysticum",
        "occultum", "omnipotentis", "philosophia", "praestantissimum",
        "quintessentia", "regeneratio", "universalis"
    ]
    prefix = random.choice(prefixes)
    suffix = random.choice(suffixes)
    parts = [prefix, suffix]
    if add_random_digit:
        random_digit = ''.join(
            random.choices(string.digits, k=digit_count)
        )
        parts.append(random_digit)
    return separator.join(parts)



from typing import Any, Mapping, List, Union
from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller, Server
from core.repo import SearchFilter
from module.auth.component.model.user_model import user_model
from module.auth.schema.user import UserData, UserLogin
from module.auth.schema.token import AccessTokenData


def register_rpc(
    logger: Logger,
    rpc_server: Server,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "auth.user"')

    @rpc_server.register('auth_user_is_admin')
    async def user_is_admin(id: str) -> bool:
        '''
        Used by RPC Authenticator
        '''
        return await user_model.is_admin(id)

    @rpc_server.register('auth_user_is_guest')
    async def user_is_guest(id: str) -> bool:
        '''
        Used by RPC Authenticator
        '''
        return await user_model.is_guest(id)

    @rpc_server.register('auth_is_user_authorized')
    async def is_user_having_permission(
        id: str, permission_name: Union[str, List[str]]
    ) -> Mapping[str, bool]:
        '''
        Used by RPC Authenticator
        '''
        if isinstance(permission_name, str):
            return await user_model.is_authorized(id, permission_name)
        return await user_model.is_authorized(id, *permission_name)

    @rpc_server.register('auth_create_token')
    async def create_token(login_data: Mapping[str, str]) -> Mapping[str, str]:
        result = await user_model.create_auth_token(UserLogin(**login_data))
        return result.dict()

    @rpc_server.register('auth_refresh_token')
    async def refresh_token(
        refresh_token: str, access_token: str
    ) -> Mapping[str, str]:
        result = await user_model.refresh_auth_token(
            refresh_token, access_token
        )
        return result.dict()

    @rpc_server.register('auth_get_user')
    async def get(
        keyword: str,
        criterion: Mapping[str, Any],
        limit: int,
        offset: int,
        user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        result = await user_model.get(
            search_filter=SearchFilter(
                keyword=keyword, criterion=criterion
            ),
            limit=limit,
            offset=offset
        )
        return result.dict()

    @rpc_server.register('auth_get_user_by_id')
    async def get_by_id(
        id: str,
        user_token_data: Mapping[str, Any] = {}
    ) -> Mapping[str, Any]:
        row = await user_model.get_by_id(id)
        return row.dict()

    @rpc_server.register('auth_insert_user')
    async def insert(
        data: Mapping[str, Any],
        user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data['created_by'] = user_token_data.user_id
        data['updated_by'] = user_token_data.user_id
        row = await user_model.insert(
            data=UserData(**data)
        )
        return row.dict()

    @rpc_server.register('auth_update_user')
    async def update(
        id: str,
        data: Mapping[str, Any],
        user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data['updated_by'] = user_token_data.user_id
        row = await user_model.update(
            id=id, data=UserData(**data)
        )
        return row.dict()

    @rpc_server.register('auth_delete_user')
    async def delete(
        id: str,
        user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        row = await user_model.delete(id=id)
        return row.dict()



from zrb.helper.codemod.add_argument_to_function_call import (
    add_argument_to_function_call
)


def test_add_argument_to_function():
    code = add_argument_to_function_call(
        code='\n'.join([
            'def fn(a, b):',
            '    pass',
            'def add(a, b, c):',
            '    pass',
            'x = fn(1, 2)',
            'y = add(1, 2)',
        ]),
        function_name='add',
        argument_name='c',
    )
    assert code == '\n'.join([
        'def fn(a, b):',
        '    pass',
        'def add(a, b, c):',
        '    pass',
        'x = fn(1, 2)',
        'y = add(1, 2, c)',
    ])



from component.log import logger
from component.db_connection import engine
from module.snake_module_name.entity.snake_entity_name.repo import (
    PascalEntityNameRepo, PascalEntityNameDBRepo
)

snake_entity_name_repo: PascalEntityNameRepo = PascalEntityNameDBRepo(
    logger=logger, engine=engine
)



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task.cmd_task import CmdTask
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from ....helper import util
from ....helper.codemod.add_import_module import add_import_module
from ....helper.codemod.append_code_to_function import append_code_to_function
from ....helper.file.text import read_text_file_async, write_text_file_async
from .._common.input import (
    project_dir_input, app_name_input, module_name_input, entity_name_input,
    plural_entity_name_input, main_column_name_input
)
from .._common.helper import validate_project_dir
from .add_navigation import create_add_navigation_task

import asyncio
import os

current_dir = os.path.dirname(__file__)
codemod_dir = os.path.join(current_dir, 'nodejs', 'codemod')


@python_task(
    name='validate',
    inputs=[
        project_dir_input, app_name_input, module_name_input, entity_name_input
    ],
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    entity_name = kwargs.get('entity_name')
    snake_module_name = util.to_snake_case(module_name)
    snake_entity_name = util.to_snake_case(entity_name)
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(
            f'Automation directory does not exist: {automation_dir}'
        )
    app_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(app_name)}'
    )
    if not os.path.exists(app_dir):
        raise Exception(f'App directory does not exist: {app_dir}')
    module_path = os.path.join(app_dir, 'src', 'module', snake_module_name)
    if not os.path.exists(module_path):
        raise Exception(f'Module directory does not exist: {module_path}')
    entity_path = os.path.join(module_path, 'entity', snake_entity_name)
    if os.path.exists(entity_path):
        raise Exception(f'Entity directory already exists: {entity_path}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        plural_entity_name_input,
        main_column_name_input,
    ],
    upstreams=[validate],
    replacements={
        'appName': '{{input.app_name}}',
        'moduleName': '{{input.module_name}}',
        'entityName': '{{input.entity_name}}',
        'pluralEntityName': '{{input.plural_entity_name}}',
        'columnName': '{{input.column_name}}',
    },
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    excludes=[
        '*/__pycache__',
    ]
)


@python_task(
    name='register-crud',
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        plural_entity_name_input,
        main_column_name_input,
    ],
    upstreams=[
        copy_resource,
    ]
)
async def register_crud(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    entity_name = kwargs.get('entity_name')
    kebab_app_name = util.to_kebab_case(app_name)
    snake_module_name = util.to_snake_case(module_name)
    snake_entity_name = util.to_snake_case(entity_name)
    await asyncio.gather(
        asyncio.create_task(register_api(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name
        )),
        asyncio.create_task(register_rpc(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name
        )),
        asyncio.create_task(register_permission(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name
        )),
    )


prepare_codemod = CmdTask(
    name='prepare-codemod',
    cwd=codemod_dir,
    cmd=[
        'npm install --save-dev',
        'node_modules/.bin/tsc'
    ]
)

add_navigation = create_add_navigation_task(
    upstreams=[
        copy_resource,
        prepare_codemod,
    ]
)


@python_task(
    name='fastapp-crud',
    group=project_add_group,
    upstreams=[
        register_crud,
        add_navigation
    ],
    runner=runner
)
async def add_fastapp_crud(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')


async def register_api(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str
):
    module_api_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'api.py'
    )
    register_function_path = '.'.join([
        'module', snake_module_name, 'entity', snake_entity_name, 'api'
    ])
    register_function = f'register_{snake_entity_name}_api'
    task.print_out(f'Read code from: {module_api_file_path}')
    code = await read_text_file_async(module_api_file_path)
    task.print_out(
        f'Add import "register_api" as "{register_function}" ' +
        f'from "{register_function_path}" to the code'
    )
    code = add_import_module(
        code=code,
        module_path=register_function_path,
        resource='register_api',
        alias=register_function
    )
    task.print_out(f'Add "{register_function}" call to the code')
    code = append_code_to_function(
        code=code,
        function_name='register_api',
        new_code=f'{register_function}(logger, app, authorizer, rpc_caller, publisher)' # noqa
    )
    task.print_out(f'Write modified code to: {module_api_file_path}')
    await write_text_file_async(module_api_file_path, code)


async def register_rpc(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str
):
    module_rpc_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'rpc.py'
    )
    register_function_path = '.'.join([
        'module', snake_module_name, 'entity', snake_entity_name, 'rpc'
    ])
    register_function = f'register_{snake_entity_name}_rpc'
    task.print_out(f'Read code from: {module_rpc_file_path}')
    code = await read_text_file_async(module_rpc_file_path)
    task.print_out(
        f'Add import "register_rpc" as "{register_function}" ' +
        f'from "{register_function_path}" to the code'
    )
    code = add_import_module(
        code=code,
        module_path=register_function_path,
        resource='register_rpc',
        alias=register_function
    )
    task.print_out(f'Add "{register_function}" call to the code')
    code = append_code_to_function(
        code=code,
        function_name='register_rpc',
        new_code=f'{register_function}(logger, rpc_server, rpc_caller, publisher)' # noqa
    )
    task.print_out(f'Write modified code to: {module_rpc_file_path}')
    await write_text_file_async(module_rpc_file_path, code)


async def register_permission(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str
):
    module_register_permission_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', 'auth',
        'register_permission.py'
    )
    task.print_out(f'Read code from: {module_register_permission_file_path}')
    code = await read_text_file_async(module_register_permission_file_path)
    code = append_code_to_function(
        code=code,
        function_name='register_permission',
        new_code='\n'.join([
            'await ensure_entity_permission(',
            f"    module_name='{snake_module_name}', entity_name='{snake_entity_name}'", # noqa
            ')'
        ])
    )
    task.print_out(
        f'Add "ensure_entity_permission" call for {snake_entity_name} ' +
        'to the code'
    )
    task.print_out(
        f'Write modified code to: {module_register_permission_file_path}'
    )
    await write_text_file_async(module_register_permission_file_path, code)



from typing import List
from zrb import CmdTask, Env, EnvFile, runner
from zrb.builtin._group import project_group
from .image import push_snake_app_name_image
from ._common import (
    CURRENT_DIR, DEPLOYMENT_DIR, TEMPLATE_ENV_FILE_NAME,
    image_input, pulumi_stack_input, replica_input, image_env,
    deployment_replica_env, pulumi_backend_url_env,
    pulumi_config_passphrase_env
)
import os

deployment_app_env_file = EnvFile(
    env_file=TEMPLATE_ENV_FILE_NAME, prefix='DEPLOYMENT_APP_ENV_PREFIX'
)

deployment_envs: List[Env] = [
    pulumi_backend_url_env,
    pulumi_config_passphrase_env,
    image_env,
    deployment_replica_env,
]

###############################################################################
# Task Definitions
###############################################################################

deploy_snake_app_name = CmdTask(
    icon='🚧',
    name='deploy-kebab-app-name',
    description='Deploy human readable app name',
    group=project_group,
    inputs=[
        image_input,
        replica_input,
        pulumi_stack_input,
    ],
    upstreams=[push_snake_app_name_image],
    cwd=DEPLOYMENT_DIR,
    env_files=[deployment_app_env_file],
    envs=deployment_envs,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'pulumi-up.sh'),
)
runner.register(deploy_snake_app_name)

destroy_snake_app_name = CmdTask(
    icon='💨',
    name='destroy-kebab-app-name',
    description='Remove human readable app name deployment',
    group=project_group,
    inputs=[
        pulumi_stack_input,
    ],
    cwd=DEPLOYMENT_DIR,
    env_files=[deployment_app_env_file],
    envs=deployment_envs,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'pulumi-destroy.sh'),
)
runner.register(destroy_snake_app_name)



from zrb.task.decorator import python_task


def test_task_with_decorated_runner():
    @python_task(
        name='hello'
    )
    async def hello(*args, **kwargs) -> str:
        return 'hello'

    main_loop = hello.create_main_loop()
    result = main_loop()
    assert result == 'hello'



from typing import Optional
from .accessories.color import colored
from .log import logger
import os


def inject_default_env():
    if 'PYTHONUNBUFFERED' not in os.environ:
        logger.info(colored('Set PYTHONUNBUFFERED to 1', attrs=['dark']))
        os.environ['PYTHONUNBUFFERED'] = '1'

    zrb_home_dir = os.path.dirname(os.path.dirname(__file__))
    if 'ZRB_HOME_DIR' not in os.environ:
        logger.info(colored(
            f'Set ZRB_HOME_DIR to {zrb_home_dir}', attrs=['dark']
        ))
        os.environ['ZRB_HOME_DIR'] = zrb_home_dir

    logger.info(colored('Getting project directory', attrs=['dark']))
    current_dir = os.getcwd()
    zrb_project_dir = _get_project_dir(current_dir)
    if zrb_project_dir is None:
        zrb_project_dir = current_dir
    logger.info(colored(
        f'Set ZRB_PROJECT_DIR to {zrb_project_dir}', attrs=['dark']
    ))
    os.environ['ZRB_PROJECT_DIR'] = zrb_project_dir

    if 'ZRB_PROJECT_NAME' not in os.environ:
        zrb_project_name = os.path.basename(zrb_project_dir)
        logger.info(colored(
            f'Set ZRB_PROJECT_NAME to {zrb_project_name}', attrs=['dark']
        ))
        os.environ['ZRB_PROJECT_NAME'] = zrb_project_name


def _get_project_dir(project_dir: str) -> Optional[str]:
    project_script = os.path.join(project_dir, 'zrb_init.py')
    if os.path.isfile(project_script):
        return project_dir
    # zrb_init.py not found, look for it on the parent directory
    parent_project_dir = os.path.dirname(project_dir)
    if parent_project_dir == project_dir:
        return None
    return _get_project_dir(parent_project_dir)



from . import base64
from . import env
from . import eval
from . import md5
from . import principle
from . import project
from . import ubuntu
from . import update
from . import version
from .devtool import devtool_install
from .generator.project import create as generate_project
from .generator.cmd_task import add as generate_cmd_task
from .generator.docker_compose_task import add as generate_docker_compose_task
from .generator.python_task import add as generate_python_task
from .generator.simple_python_app import add as generate_simple_python_app
from .generator.fastapp import add as generate_fastapp
from .generator.fastapp_module import add as generate_fastapp_module
from .generator.fastapp_crud import add as generate_fastapp_crud
from .generator.fastapp_field import add as generate_fastapp_field

assert base64
assert env
assert eval
assert md5
assert principle
assert project
assert ubuntu
assert update
assert version
assert devtool_install
assert generate_project
assert generate_cmd_task
assert generate_docker_compose_task
assert generate_python_task
assert generate_simple_python_app
assert generate_fastapp
assert generate_fastapp_module
assert generate_fastapp_crud
assert generate_fastapp_field



from .runner import runner
from .helper.loader.load_module import load_module
from .task.decorator import python_task
from .task.task import Task
from .task.cmd_task import CmdTask
from .task.docker_compose_task import DockerComposeTask
from .task.http_checker import HTTPChecker
from .task.port_checker import PortChecker
from .task.path_checker import PathChecker
from .task.resource_maker import ResourceMaker, Replacement, ReplacementMutator
from .task_input.input import Input
from .task_input.bool_input import BoolInput
from .task_input.choice_input import ChoiceInput
from .task_input.float_input import FloatInput
from .task_input.int_input import IntInput
from .task_input.password_input import PasswordInput
from .task_input.str_input import StrInput
from .task_env.env import Env
from .task_env.env_file import EnvFile
from .task.base_task import Group
from .builtin import _group as builtin_group
from .helper.default_env import inject_default_env


assert runner
assert load_module
assert python_task
assert Task
assert CmdTask
assert DockerComposeTask
assert HTTPChecker
assert PortChecker
assert PathChecker
assert ResourceMaker
assert Replacement
assert ReplacementMutator
assert Input
assert BoolInput
assert ChoiceInput
assert FloatInput
assert IntInput
assert PasswordInput
assert StrInput
assert Env
assert EnvFile
assert Group
assert builtin_group

inject_default_env()



from module.auth.core import (
    AccessTokenScheme, create_oauth2_bearer_access_token_scheme
)
from module.auth.component.access_token_util import access_token_util
from module.auth.component.user import guest_user
from config import app_auth_access_token_cookie_key


access_token_scheme: AccessTokenScheme = create_oauth2_bearer_access_token_scheme( # noqa
    guest_user=guest_user,
    access_token_util=access_token_util,
    token_url='/api/v1/auth/login-oauth',
    token_cookie_key=app_auth_access_token_cookie_key
)



from pulumi_kubernetes.helm.v3 import Chart, LocalChartOpts
from _common import (
    NAMESPACE, REDPANDA_AUTH_MECHANISM, REDPANDA_AUTH_SASL_ENABLED,
    REDPANDA_AUTH_USER_NAME, REDPANDA_AUTH_USER_PASSWORD
)


def create_redpanda() -> Chart:
    redpanda = Chart(
        'redpanda',
        LocalChartOpts(
            path='./helm-charts/redpanda',
            namespace=NAMESPACE,
            values={
                'auth': {
                    'sasl': {
                        'enabled': REDPANDA_AUTH_SASL_ENABLED,
                        'mechanism': REDPANDA_AUTH_MECHANISM,
                        'users': [
                            {
                                'name': REDPANDA_AUTH_USER_NAME,
                                'password': REDPANDA_AUTH_USER_PASSWORD,
                                'mechanism': REDPANDA_AUTH_MECHANISM
                            }
                        ]
                    }
                },
            },
            skip_await=True
        ),
    )
    return redpanda
